""" Demo puller module """
from typing import List
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.demo.demo_passenger_1 import DemoPassenger1
from databus.passenger.attachment import Attachment, AttachmentFormat
from databus.puller.abstract_puller import AbstractPuller


class DemoPuller3(AbstractPuller):
    """ Demo puller class """

    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Fake operation """
        for seated_passenger in p_seated_passengers:
            self.log.append_text(
                f"Demo puller 3 notified about seated passenger {seated_passenger.id_text}")

    def pull(self) -> List[DemoPassenger1]:
        """ Fake operation """
        output = []

        passenger1 = DemoPassenger1()
        passenger1.external_id = "ID_3_1"
        passenger1.dataset = "Puller 3 pulled first DemoPassenger1"
        passenger1.source_system = "DEMO_SYSTEM"
        passenger1.puller_module = self.__module__
        passenger1.attachments.append(Attachment(p_name="puller3_file1.txt",
                                                 p_format=AttachmentFormat.text,
                                                 p_text_content="Lorem Ipsum"))
        output.append(passenger1)
        self.log.append_text(f"Got passenger {passenger1.id_text}")

        return output



""" Unzip module """
from typing import List
import mimetypes
import shutil
import os
from zipfile import ZipFile
from databus.passenger.attachment import Attachment, AttachmentFormat

class Unzipper:
    """ Main unzip class """
    _TMP_ZIP_DIR = "_tmp_zip"
    _TMP_ZIP_EXTRACT_DIR = "_extracted"
    _TMP_ZIP_FILE = "_tmp_zip.zip"

    def __init__(self):
        self._mime_type = mimetypes.MimeTypes()
        self._attachments = []
        self._deletable_indices = []
        self._new_attachments = []
        self._eligible_file_extensions = []
        self._extract_dir = os.path.join(Unzipper._TMP_ZIP_DIR, Unzipper._TMP_ZIP_EXTRACT_DIR)
        self._zip_path = os.path.join(Unzipper._TMP_ZIP_DIR, Unzipper._TMP_ZIP_FILE)

    def execute(self,
                p_attachments: List[Attachment],
                p_extensions: List[str] = None):
        """ Unzips attachments """
        self._attachments = p_attachments
        self._deletable_indices = []
        self._new_attachments = []

        if p_extensions is None:
            self._eligible_file_extensions = []
        else:
            self._eligible_file_extensions = p_extensions

        self._unzip()
        self._delete_zip_files()
        self._append_unzipped_files()

    def _download_zip_file(self, p_bin):
        shutil.rmtree(Unzipper._TMP_ZIP_DIR, ignore_errors=True)
        os.makedirs(Unzipper._TMP_ZIP_DIR, exist_ok=True)
        with open(self._zip_path, "wb") as zip_file:
            zip_file.write(p_bin)

    def _is_file_eligible(self, p_filename: str) -> bool:
        if p_filename is None or len(p_filename) <= 0:
            return False
        if len(self._eligible_file_extensions) <= 0:
            return True
        filename_len = len(p_filename)
        for efe in self._eligible_file_extensions:
            extension_len = len(efe)
            if filename_len <= (extension_len+1):
                continue
            filename_fragment = "." + p_filename[-extension_len:]
            supposed_extension = "." + efe
            if filename_fragment.lower() == supposed_extension.lower():
                return True

        return False

    def _extract_zip_file(self):
        extracted_files = []
        with ZipFile(self._zip_path) as zip_file:
            files_in_zip = zip_file.namelist()
            for file_in_zip in files_in_zip:
                if not self._is_file_eligible(file_in_zip):
                    continue
                extract_path = os.path.join(self._extract_dir, file_in_zip)
                zip_file.extract(file_in_zip, self._extract_dir)
                extracted_files.append(extract_path)

        for extracted_path in extracted_files:
            mime = self._mime_type.guess_type(extracted_path)[0]
            file_format = Attachment.guess_format_by_mime_type(mime)
            file_in_zip = os.path.split(extracted_path)[1]

            if file_format == AttachmentFormat.text:
                with open(extracted_path, "r", encoding="utf-8") as extracted_file:
                    file_content = extracted_file.read()
                    file_content = file_content.replace("\n", " ")
                    file_content = file_content.replace("\t", " ")
                unzip_attachment = Attachment(
                    p_name=file_in_zip,
                    p_format=AttachmentFormat.text,
                    p_text_content=file_content)
            else:
                with open(extracted_path, "rb") as extracted_file:
                    file_content = extracted_file.read()
                unzip_attachment = Attachment(
                    p_name=file_in_zip,
                    p_format=AttachmentFormat.binary,
                    p_binary_content=file_content)

            self._new_attachments.append(unzip_attachment)

    def _unzip(self):
        attachment_index = -1

        for attachment in self._attachments:
            attachment_index += 1
            if len(attachment.name) < 4:
                continue
            if attachment.name.upper()[-4:] != ".ZIP":
                continue

            self._download_zip_file(attachment.binary_content)
            self._extract_zip_file()
            self._deletable_indices.append(attachment_index)
            shutil.rmtree(Unzipper._TMP_ZIP_DIR, ignore_errors=True)

    def _delete_zip_files(self):
        self._deletable_indices.sort(reverse=True)
        for deletable_index in self._deletable_indices:
            self._attachments.pop(deletable_index)

    def _append_unzipped_files(self):
        for new_attachment in self._new_attachments:
            self._attachments.append(new_attachment)



""" Abstract passenger module """
from abc import ABC
from datetime import datetime
from typing import List
from uuid import uuid1, UUID
from databus.passenger.attachment import Attachment
from databus.passenger.unzip import Unzipper
from databus.passenger.html_link_downloader import HtmlLinkDownloader

class AbstractPassenger(ABC): # pylint: disable=R0903, R0902
    """ Abstract passenger class """
    _clock_seq: int = 0

    def __init__(self, # pylint: disable=R0913, R0912
                 p_external_id: str = None,
                 p_internal_id: UUID = None,
                 p_source_system: str = None,
                 p_attachments: List[Attachment] = None,
                 p_puller_module: str = None,
                 p_pull_datetime: datetime = None,
                 p_passenger_module: str = None,
                 p_log_guids: List[UUID] = None):

        if p_external_id is None:
            self.external_id = ""
        else:
            self.external_id = p_external_id

        if p_source_system is None:
            self.source_system = ""
        else:
            self.source_system = p_source_system

        if p_puller_module is None:
            self.puller_module = ""
        else:
            self.puller_module = p_puller_module

        if p_pull_datetime is None:
            self.pull_datetime = datetime.now()
        else:
            self.pull_datetime = p_pull_datetime

        if p_attachments is None:
            self.attachments = []
        else:
            self.attachments = p_attachments

        if p_internal_id is None:
            AbstractPassenger._clock_seq += 1
            self.internal_id = uuid1(clock_seq=AbstractPassenger._clock_seq)
        else:
            self.internal_id = p_internal_id

        if p_passenger_module is None:
            self.passenger_module = self.__module__
        else:
            self.passenger_module = p_passenger_module

        if p_log_guids is None:
            self._log_guids = []
        else:
            self._log_guids = p_log_guids

    @property
    def id_text(self) -> str:
        """ Returns the unique ID of the passenger as text """
        return self.source_system + " - " + self.external_id + " (" + str(self.internal_id) + ")"

    @property
    def log_guids(self) -> List[UUID]:
        """ Returns log guids """
        return self._log_guids

    def collect_log_guid(self, guid: UUID):
        """ Appends a new log guid, preventing duplicates """
        if guid in self._log_guids:
            return
        self._log_guids.append(guid)

    def get_attachment_by_name(self, name: str) -> Attachment:
        """ Returns the attachment """
        for attachment in self.attachments:
            if attachment.name == name:
                return attachment
        return None

    def unzip_attachments(self, p_extensions: List[str] = None):
        """ Finds .ZIP attachments, and turns them into regular
        attachments within the same object
        """
        Unzipper().execute(self.attachments, p_extensions)

    def download_links_in_html_as_attachments(self, p_html: str, p_extensions: List[str]):
        """ Scans the given HTML file, finds links, downloads
        the files and saves them as attachments.
        This method supports only text attachments at this time.
        """
        HtmlLinkDownloader().execute(self.attachments, p_html, p_extensions)



""" Implementation module for JSON database """
from datetime import datetime
from shutil import disk_usage
from typing import List
from uuid import UUID
from databus.client.client import Client
from databus.client.log import Log, LogEntry, MessageType
from databus.client.user import Credential
from databus.client.customizing import ClientCustomizing
from databus.database.abstract_database import AbstractDatabase, LogListItem
from databus.database.json_db.json_client import JsonClient
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_log import JsonLog
from databus.database.json_db.json_queue import JsonQueue
from databus.database.json_db.json_cust_node_backup import CustomizingNodeBackup
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.pqueue.queue_status import QueueStatus, PassengerQueueStatus

ARGS_TEMPLATE = JsonDatabaseArguments.TEMPLATE

class JsonDatabase(AbstractDatabase):
    """ Implementation class for JSON database """
    _FREE_SPACE_ERROR_PERC = 5

    def __init__(self,
                 p_client_id: str,
                 p_log: Log,
                 p_passenger_factory: AbstractPassengerFactory,
                 p_arguments: dict):
        super().__init__(p_client_id, p_log, p_passenger_factory, p_arguments)
        self._args = JsonDatabaseArguments(p_arguments)
        self._json_client = JsonClient(self._args)
        if p_client_id is None or p_client_id == "":
            self.client = None
        else:
            try:
                self.client = self._get_client(p_client_id)
            except Exception: # pylint: disable=W0703
                self.client = None # We might be creating a new database
        self._json_log = JsonLog(self._args)
        self._json_queue = JsonQueue(p_client_id, p_log, self.passenger_factory, self._args)
        self._cust_node_backup = CustomizingNodeBackup(json_db_arguments=self._args)
        self._check_disk_size(p_log)

    @property
    def client_master_data(self) -> str:
        """ Returns the definition of all clients as a JSON string """
        return self._json_client.client_master_as_json

    @client_master_data.setter
    def client_master_data(self, p_definitions: str):
        """ Sets the definitions into the clients config JSON """
        self._json_client.client_master_as_json = p_definitions

    @property
    def customizing(self) -> str:
        """ Returns the client customizing as JSON """
        return self._json_client.get_config_as_string(self.client_id)

    @customizing.setter
    def customizing(self, p_customizing: str):
        """ Sets the customizing into the clients config JSON """
        self._json_client.save_config(self.client_id, p_customizing)

    def delete_old_logs(self, p_before: datetime):
        """ Deletes old logs from the disk """
        self.log.append_text(f"Deleting logs before {p_before.isoformat()}")
        self._json_log.delete_log_file_before(self.client.id, p_before, self.log)

    def delete_passenger_queue(self, p_passengers: List[AbstractPassenger]):
        """ Deletes passenger from disk """
        self.log.append_text("Deleting passengers from queue")
        self._json_queue.delete_passengers(p_passengers)

    def ensure_schema_existence(self):
        """ Checks the schema for the client
        Creates / completes the schema if anything is missing
        """
        self._json_client.ensure_schema_existence(self.client_id)

    def erase_passenger_queue(self):
        """ Deletes all passengers from disk """
        self.log.append_text("Erasing passenger queue")
        self._json_queue.erase_passenger_queue()

    def get_clients(self) -> List[Client]:
        """ Returns a list of clients """
        return self._json_client.all_clients

    def get_log_content(self, p_log_id: str) -> str:
        """ Returns the contents of the given log
        p_log_id is whatever you have returned in get_log_list.
        """
        return self._json_log.get_log_file_content(self.client_id, p_log_id)

    def get_log_list(self) -> List[LogListItem]:
        """ Returns log entries """
        output = []
        b_warning = f"[{MessageType.warning.name}]"
        b_error = f"[{MessageType.error.name}]"

        log_ids = self._json_log.get_log_file_list(self.client_id)
        for log_id in log_ids:
            output_item = LogListItem(p_log_id=log_id)
            log_content = self.get_log_content(log_id)
            if b_error in log_content:
                output_item.worst_message_type = MessageType.error
            elif b_warning in log_content:
                output_item.worst_message_type = MessageType.warning
            else:
                output_item.worst_message_type = MessageType.info
            output.append(output_item)
        return output

    def get_passenger_queue_entries(self, # pylint: disable=R0913
                                    p_passenger_module: str = None,
                                    p_processor_status: QueueStatus = None,
                                    p_pusher_status: QueueStatus = None,
                                    p_puller_notified: bool = None,
                                    p_pulled_before: datetime = None
                                    ) -> List[PassengerQueueStatus]:
        """ Reads the requested entries from the disk """
        self.log.append_text("Reading passenger queue entries")
        return self._json_queue.get_passengers(p_passenger_module,
                                               p_processor_status,
                                               p_pusher_status,
                                               p_puller_notified,
                                               p_pulled_before)

    def get_passenger_queue_entry(self, # pylint: disable=R0913
                                  p_internal_id: str
                                 ) -> PassengerQueueStatus:
        return self._json_queue.get_passengers(p_internal_id=p_internal_id)[0]

    def insert_client(self, p_client: Client):
        """ Inserts a new client """
        self._json_client.create_client_if_missing__by_obj(p_client)

    def insert_log(self, p_log: Log):
        """ Creates a new log file on the disk """
        self.log.append_text("Writing log to disk")
        if self.client is None:
            return
        self._json_log.insert(self.client.id, p_log)

    def insert_passenger_queue(self, p_passenger_status: PassengerQueueStatus):
        """ Writes new files to the disk """
        self.log.append_text(f"Appending passenger {p_passenger_status.passenger.id_text}")
        self._json_queue.insert_passenger(p_passenger_status)

    def update_queue_status(self, p_status: PassengerQueueStatus):
        """ Updates queue files on the disk """
        self.log.append_text(f"Updating passenger {p_status.passenger.id_text}")
        self._json_queue.update_passenger(p_status)

    def update_user_credential(self, p_credential: Credential):
        """ Updates the credential of the given user """
        self._json_client.update_user_credential(self.client_id, p_credential)

    def convert_log_guid_to_id(self, p_guid: UUID) -> str:
        """ UUID to id conversion """
        guid_as_str = str(p_guid)
        log_ids = self._json_log.get_log_file_list(self.client_id)
        for log_id in log_ids:
            if guid_as_str in log_id:
                return log_id
        return ""

    def backup_client_customizing(self, p_cc: ClientCustomizing):
        """ Backup customizing nodes """
        self._cust_node_backup.execute(p_cc=p_cc)

    def delete_old_client_customizing_backups(self, p_before: datetime, p_log: Log):
        """ Deletes overdue client customizing backups """
        self._cust_node_backup.delete_old_files(p_before=p_before, p_db=self, p_log=p_log)

    def _get_client(self, p_id: str) -> Client:
        return self._json_client.get_single(p_id)

    def _check_disk_size(self, p_log: Log):
        total, used, free = disk_usage("/")
        divider = 2**30
        total_gb = total // divider
        used_gb = used // divider
        free_gb = free // divider
        free_perc = int((free / total) * 100)

        p_log.append_text(f"Disk {total_gb}GB, used {used_gb}GB, free {free_gb}GB ({free_perc})%")

        if free_perc <= JsonDatabase._FREE_SPACE_ERROR_PERC:
            size_entry = LogEntry(p_message="Disk space running low", p_type=MessageType.error)
            p_log.append_entry(size_entry)



""" Abstract processor factory module """
from enum import Enum
from typing import Protocol
from databus.client.log import Log
from databus.processor.abstract_processor import AbstractProcessor

class ProcessorCreationError(Exception):
    """ Abstract processor creation exception """
    class ErrorCode(Enum):
        """ Error code enum """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self, p_error_code: ErrorCode, p_module: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Error message as text """
        if self.error_code == ProcessorCreationError.ErrorCode.cant_create_instance:
            return f"Can't create {self.module} processor instance"
        if self.error_code == ProcessorCreationError.ErrorCode.parameter_missing:
            return "Parameters missing, can't create processor instance"
        return "Processor creation error"


class AbstractProcessorFactory(Protocol): # pylint: disable=R0903
    """ Abstract processor factory """
    def create_processor(self, p_module: str, p_log: Log) -> AbstractProcessor:
        """ Processor factory """



""" Pause web module """
from flask import redirect, url_for, render_template
from databus.client.client import Client
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError

class PauseController(AbstractController):
    """ Pause controller """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        if self.authenticated_client_id != Client.ROOT:
            return redirect(url_for("_login"), code=302)

        try:
            self.dispatcher.request_pause()
            success = True
            error = ""
        except Exception as raised_error:
            success = False
            error = f"Pause error: {str(raised_error)}"

        return render_template(
            "pause.html",
            alias=self.dispatcher.ticket.system_alias,
            success=success,
            error=error)



""" Client customizing backup """
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher

def backup_client_customizings(p_dispatcher: AbstractDispatcher):
    """ Backup client customizing"""
    customs = p_dispatcher.get_client_customizing_list()

    for custom in customs:
        db = p_dispatcher.get_client_database(custom.client.id)
        db.backup_customizing_nodes(custom)



""" Client customizing """
from typing import List
from databus.client.client import Client

class CustomizingNode:  # pylint: disable=R0903
    """ Defines a customizing node """
    def __init__(self, p_name: str = "", p_content: str = ""):
        self.name = p_name
        self.content = p_content


class ClientCustomizing: # pylint: disable=R0903
    """ Defines a client and its customizing files """
    def __init__(self, p_client: Client = None, p_customs: List[CustomizingNode] = None):
        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client

        if p_customs is None:
            self.nodes = []
        else:
            self.nodes = p_customs



""" Module for customizing pages """
from flask import render_template, redirect, request, url_for
from databus.dispatcher.abstract_dispatcher import CLIENTS_DB_NODE, DATABUS_DB_NODE
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError

class CustomizingListController(AbstractController):
    """ Customizing list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        entries = self.dispatcher.get_client_customizing_list(
            p_client_id=self.authenticated_client_id)

        return render_template(
            "customizing_list.html",
            entries=entries,
            alias=self.dispatcher.ticket.system_alias)

class CustomizingEditController(AbstractController):
    """ Customizing list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        node = request.args.get("node", 0, type=str)
        entry = self.dispatcher.get_client_customizing_entry(self.requested_client_id, node)
        return render_template("customizing_edit.html", client=self.requested_client_id, node=node, entry=entry, alias=self.dispatcher.ticket.system_alias) # pylint: disable=C0301

class CustomizingSaveController(AbstractController):
    """ Customizing save """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        client_db = self.requested_client_database
        node = request.form["node"]
        customizing = request.form["customizing"]
        customizing = customizing.replace("\r\n", "\n") # Windows fix

        if node == DATABUS_DB_NODE:
            client_db.customizing = customizing
        elif node == CLIENTS_DB_NODE:
            client_db.client_master_data = customizing
        else:
            external_file = self.dispatcher.external_config_file_manager.get_file(self.requested_client_id, node) # pylint: disable=C0301
            external_file.file_content = customizing

        url = url_for("_customizing_list")
        url += "?" + AbstractController._get_cache_buster()
        return redirect(url, code=302)



""" Default dispatcher module """
from datetime import datetime, timedelta
from threading import Thread
from time import sleep
from typing import List
import os
import atexit
from databus.client.client import Client
from databus.client.client_passenger import ClientPassenger
from databus.client.log import Log, LogEntry, MessageType
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher, DispatcherTicket
from databus.driver.abstract_driver import BusTicket
from databus.web import app

class ClientPassengerTickCount:
    """ Keeps track of passenger ticks """
    _INFINITE_TICK = 999999999

    def __init__(self):
        self._ticks = {}

    def collect(self, p_client_id: str, p_passenger_id: str):
        """ Inserts the client & passenger if they weren't before """
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        if key not in self._ticks:
            self._ticks[key] = 0

    def collect_clients(self, p_clients: List[Client]):
        """ Inserts missing clients """
        for client in p_clients:
            for passenger in client.passengers:
                self.collect(client.id, passenger.name)

    def expedite_client_passenger(self, p_client_id: str, p_passenger_id: str):
        """ Prioritizes the passenger in the next cycle """
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        if key not in self._ticks:
            return
        self._ticks[key] = ClientPassengerTickCount._INFINITE_TICK

    def get_tick(self, p_client_id: str, p_passenger_id: str):
        """ Returns the current tick count """
        self.collect(p_client_id, p_passenger_id)
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        return self._ticks[key]

    def reset_tick(self, p_client_id: str, p_passenger_id: str):
        """ Resets tick """
        self.collect(p_client_id, p_passenger_id)
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        self._ticks[key] = 0

    def tick(self):
        """ Increases tick counts for all client passengers """
        for self_tick in self._ticks:
            self._ticks[self_tick] += 1

    @staticmethod
    def _build_key(p_client_id: str, p_passenger_id: str):
        return p_client_id + ":-:" + p_passenger_id


class DispatchState: # pylint: disable=R0903
    """ State of the dispatcher """
    def __init__(self):
        self.clients = List[Client]


class DispatcherStatus:
    """ Status of the dispatcher """
    def __init__(self):
        self.pausing = False
        self.paused = False
        self.dispatching = False
        self.exporting = False
        self.shutting_down = False


class PrimalDispatcher(AbstractDispatcher): # pylint: disable=R0903
    """ Default dispatcher implementation """

    def __init__(self, p_ticket: DispatcherTicket = None):
        super().__init__(p_ticket)
        self._dispatch_state = DispatchState()
        self._next_dispatch_time = datetime.now()
        self._tick_count = ClientPassengerTickCount()
        self._status = DispatcherStatus()
        atexit.register(self._shutdown_when_safely_possible)

    @property
    def dispatching(self) -> bool:
        """ Is the dispatcher active or not """
        return self._status.dispatching

    @property
    def paused(self) -> bool:
        """ Is the dispatcher paused or not """
        return self._status.pausing or self._status.paused

    @property
    def exporting(self) -> bool:
        """ Is export active or not """
        return self._status.exporting

    @property
    def shutting_down(self) -> bool:
        """ Is shutdown active or not """
        return self._status.shutting_down

    def expedite_client_passenger(self, p_client_id: str, p_passenger_module: str):
        """ Prioritizes the passenger in the next cycle """
        self._tick_count.expedite_client_passenger(p_client_id, p_passenger_module)

    def export_data_begin(self):
        """ Indicates that data export is starting """
        self.request_pause()
        self._wait_until_paused()
        self._status.exporting = True

    def export_data_end(self):
        """ Indicates that data export is ending """
        self._status.exporting = False
        self.resume()

    def start(self):
        """ Starts the dispatcher timer and web server """
        if self.ticket.run_web_server:
            Thread(target=self._start_web_server, daemon=True).start()

        self.backup_client_customizings()
        self._start_dispatch_timer()

    def request_shutdown(self):
        """ Dispatcher shutdown """
        self._status.shutting_down = True
        Thread(target=self._shutdown_when_safely_possible, daemon=True).start()

    def request_pause(self):
        """ Pauses dispatcher
        This is the antonym of resume
        """
        if self.paused:
            return
        if self._status.shutting_down:
            raise Exception("Can't pause while shutting down")

        self._status.pausing = True
        self._status.paused = False
        Thread(target=self._pause_when_safely_possible, daemon=True).start()

    def resume(self):
        """ Resumes the dispatcher
        This is the antonym of pause
        """
        if self._status.pausing:
            raise Exception("Pause in progress, can't resume yet")
        self._status.pausing = False
        self._status.paused = False

    def _dispatch(self):
        try:
            if self.paused or self._status.shutting_down:
                return
            self._status.dispatching = True
            self._dispatch_state = DispatchState()
            self._read_clients()
            self._drive_high_time_passengers()
        finally:
            self._status.dispatching = False

    def _drive_high_time_passengers(self):
        for client in self._dispatch_state.clients:
            for client_passenger in client.passengers:
                try:
                    tick_count = self._tick_count.get_tick(client.id, client_passenger.name)
                    if tick_count < client_passenger.sync_frequency:
                        continue
                    self._tick_count.reset_tick(client.id, client_passenger.name)
                    self._drive_passenger(client, client_passenger)
                except Exception as drive_error: # pylint: disable=W0703
                    print(str(drive_error))

    def _drive_passenger(self, p_client: Client, p_client_passenger: ClientPassenger):
        if self.paused or self._status.shutting_down:
            return
        log = Log()
        db = None # pylint: disable=C0103
        driver = None
        try:
            log.append_text("Dispatching client " +
                            p_client.id +
                            " passenger " +
                            p_client_passenger.name)

            log.append_text(f"Creating database {self.ticket.database_module}")
            db = self.get_client_database(p_client.id, log) # pylint: disable=C0103
            log.append_text("Checking schema")
            db.ensure_schema_existence()
            log.append_text(f"Creating driver {self.ticket.driver_module}")
            driver = self.get_driver()
            log.append_text("Creating ticket")

            ticket = BusTicket(
                p_client_passenger=p_client_passenger,
                p_log=log,
                p_database=db)

            log.append_text("Driving")
            driver.drive(ticket)

        except Exception as drive_error: # pylint: disable=W0703
            if log is not None:
                log.append_entry(LogEntry(p_message=str(drive_error), p_type=MessageType.error))
        finally:
            self._tick_count.reset_tick(p_client.id, p_client_passenger.name)
            if self.ticket.dispatcher_observer is not None and log is not None:
                self.ticket.dispatcher_observer.drive_passenger_complete(p_client,
                                                                         p_client_passenger,
                                                                         log)
            if db is not None:
                db.insert_log(log)
                db.delete_old_logs(p_client.log_expiry_date)
                db.delete_old_client_customizing_backups(
                    p_before=p_client.log_expiry_date,
                    p_log=log)
            if driver is not None:
                driver.queue.delete_completed_passengers(
                    p_client_passenger.name,
                    p_client_passenger.queue_expiry_date)

    def _read_clients(self):
        self._dispatch_state.clients = self.all_clients
        self._tick_count.collect_clients(self._dispatch_state.clients)
        self._tick_count.tick()

    def _sleep_until_next_dispatch_time(self):
        now = datetime.now()
        if now >= self._next_dispatch_time:
            return
        seconds_to_sleep = (self._next_dispatch_time - now).seconds
        sleep(seconds_to_sleep)

    def _start_dispatch_timer(self):
        while True:
            self._next_dispatch_time = self._next_dispatch_time + timedelta(0, 60)
            self._dispatch()
            if self._status.shutting_down:
                return
            self._sleep_until_next_dispatch_time()

    def _start_web_server(self):
        app.run_web_server(self)

    def _pause_when_safely_possible(self):
        while True:
            if not self._status.dispatching:
                self._status.pausing = False
                self._status.paused = True
                return
            sleep(1)

    def _wait_until_paused(self):
        while True:
            if self.paused:
                return
            if (not self._status.pausing) and (not self._status.paused):
                raise Exception("Pause failed")
            sleep(1)

    def _shutdown_when_safely_possible(self):
        while True:
            if not self._status.dispatching and not self._status.exporting:
                os._exit(0) # pylint: disable=W0212
                return
            sleep(1)



""" Customizing node backup """
from datetime import datetime
from os import listdir, path, remove
import uuid
import pathlib
from databus.client.customizing import ClientCustomizing
from databus.client.log import Log, LogEntry, MessageType
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_path_builder import JsonPathBuilder
from databus.database.abstract_database import AbstractDatabase

class CustomizingNodeBackup:
    """ Customizing node backup """
    def __init__(self, json_db_arguments: JsonDatabaseArguments):
        self._args = json_db_arguments

    def execute(self, p_cc: ClientCustomizing):
        """ Executes main operation """
        path_builder = JsonPathBuilder(p_cc.client.id, p_args=self._args)
        pathlib.Path(path_builder.backup_root_path).mkdir(parents=True, exist_ok=True)
        backup_id = str(uuid.uuid4())

        for node in p_cc.nodes:
            node_file = f"{backup_id}-{node.name}.json"
            node_path = path.join(path_builder.backup_root_path, node_file)
            with open(node_path, "w", encoding="utf-8") as open_node_path:
                open_node_path.write(node.content)

    def delete_old_files(self, p_before: datetime, p_db: AbstractDatabase, p_log: Log):
        """ Deletes expired backup files """
        clients = p_db.get_clients()

        for client in clients:
            path_builder = JsonPathBuilder(client.id, p_args=self._args)
            bak_path = path_builder.backup_root_path
            bak_files = [f for f in listdir(bak_path) if path.isfile(path.join(bak_path, f))]

            for bak_file in bak_files:
                try:
                    bak_file_path = path.join(bak_path, bak_file)
                    create_ctime = path.getctime(bak_file_path)
                    create_dt = datetime.fromtimestamp(create_ctime)

                    if create_dt > p_before:
                        continue

                    remove(bak_file_path)

                except Exception as error:
                    if p_log is None:
                        continue
                    log_entry = LogEntry(p_message=str(error), p_type=MessageType.error)
                    p_log.append_entry(log_entry)



""" Shutdown web module """
from flask import redirect, url_for, render_template
from databus.client.client import Client
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class SystemController(AbstractController):
    """ System controller """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        if self.authenticated_client_id != Client.ROOT:
            return redirect(url_for("_login"), code=302)

        if self.dispatcher.shutting_down:
            status = "Shutting down"
        elif self.dispatcher.exporting:
            status = "Exporting data"
        elif self.dispatcher.paused:
            status = "Paused"
        elif self.dispatcher.dispatching:
            status = "Dispatching"
        else:
            status = "Idle"

        return render_template(
            "system.html",
            alias=self.dispatcher.ticket.system_alias,
            status=status)



""" HTML link downloader module """
from typing import List
import os
import requests
from urlextract import URLExtract
from databus.passenger.attachment import Attachment, AttachmentFormat

class HtmlLinkDownloader:
    """ Scans the given HTML file, finds links, downloads
    the files and saves them as attachments.
    This method supports only text attachments at this time.
    """
    def __init__(self):
        self._extensions = []
        self._attachments = []

    def execute(self, p_attachments: List[Attachment], p_html: str, p_extensions: List[str]):
        """ Main method """
        self._extensions = p_extensions
        self._attachments = p_attachments
        urls = HtmlLinkDownloader._extract_urls(p_html)

        for url in urls:
            low_url = url.lower()
            if not self._has_eligible_extension(low_url):
                continue
            clean_url = HtmlLinkDownloader._clean_url(low_url)
            self._download(clean_url)

    @staticmethod
    def _extract_urls(p_html: str):
        # Build clean HTML
        if p_html is None or len(p_html) <= 0:
            return []
        clean_html = p_html.replace("\r", "").replace("\n", "")
        html_tag_pos = clean_html.lower().find("<html")
        if html_tag_pos < 0:
            return []
        clean_html = clean_html[html_tag_pos:]

       # Extract URL's
        extractor = URLExtract()
        urls = extractor.find_urls(clean_html)
        return urls

    @staticmethod
    def _clean_url(url):
        if "urldefense.com" in url:
            real_http_pos = url.rfind("http")
            clean_url = url[real_http_pos:].replace("__", "")
        else:
            clean_url = url
        if clean_url[-1] == "/":
            clean_url = clean_url[:-1]
        return clean_url

    def _has_eligible_extension(self, url):
        has_eligible_extension = False
        for extension in self._extensions:
            low_extension = "." + extension.lower()
            if low_extension in url:
                has_eligible_extension = True
                break
        return has_eligible_extension

    def _download(self, url):
        filename = os.path.basename(url)
        dummy_name, extension = os.path.splitext(filename)
        extension = extension.replace(".", "")
        file_format = Attachment.guess_format_by_file_extension(extension)

        response = requests.get(url, allow_redirects=True)

        if file_format == AttachmentFormat.text:
            downloaded_attachment = Attachment(
                p_name=filename,
                p_format=AttachmentFormat.text,
                p_text_content=response.text)
        else:
            downloaded_attachment = Attachment(
                p_name=filename,
                p_format=AttachmentFormat.binary,
                p_binary_content=response.content)

        self._attachments.append(downloaded_attachment)
